#define STRICT
#include <windows.h>
#include <windowsx.h>
#include <assert.h>
#include <process.h>

/* ========================================================================== */

#define XM_NOTIFY        0x3000
#define XM_THREADPARAM   0x3001

#define ID_TILEVERT      (-100)
#define ID_TILEHORZ      (-101)
#define ID_CASCADE       (-102)
#define ID_ARRANGE       (-103)

/* ========================================================================== */

class CLASSINFO
{
  protected:

  HINSTANCE hInst;
  ATOM      ClassAtom;

  public:

            CLASSINFO (HINSTANCE h)
            {
              hInst     = h;
              ClassAtom = INVALID_ATOM;
            };

           ~CLASSINFO (void)
            {
              if (ClassAtom != INVALID_ATOM)
                UnregisterClass (MAKEINTATOM (ClassAtom), hInst);
            };

  BOOL OK (void) const
  {
    return ClassAtom != INVALID_ATOM;
  };

  LPCSTR ClassName (void) const
  {
    return MAKEINTATOM (ClassAtom);
  };

  HINSTANCE Instance (void) const
  {
    return hInst;
  };
};

/* ========================================================================== */

class FRAMECLASSINFO : public CLASSINFO
{
  public:

          FRAMECLASSINFO (HINSTANCE h);
};

/* ========================================================================== */

// MDI child window class

class CHILDCLASSINFO : public CLASSINFO
{
  public:

          CHILDCLASSINFO (HINSTANCE h);
};

/* ========================================================================== */

// MDI Frame handler class

class FrameCode
{
  friend class ChildCode;

  FRAMECLASSINFO frameclass;
  CHILDCLASSINFO childclass;
  HWND           hFrame;
  HWND           hClient;
  HMENU          hMenu;

  public:

                           FrameCode (HINSTANCE hInst);
                          ~FrameCode (void) {};

          int              Run           (void);
          void             CreateMenu    (void);
          void             RefreshMenu   (void) const;
          void             OnCommand     (WPARAM);
          void             OnClose       (void);
          BOOL             CloseChildren (void);
          void             Bind          (HWND);
          void             UnBind        (void);
  static  FrameCode *      Object        (HWND);

  // Test functions

          void             StartChild (void);
          void             RunChild   (int);
          void             Maximize   (void);

  // Frame window callback

  static  LRESULT CALLBACK MDIFrameProc (HWND, UINT, WPARAM, LPARAM);
};

/* ========================================================================== */

// MDI Child window handler class

class ChildCode
{
  FrameCode *     frame;
  HWND            handle;

  public:

                           ChildCode (FrameCode *);
                          ~ChildCode (void) { assert (frame == NULL); };

          void             Create    (int);
          void             Bind      (HWND);
          void             UnBind    (void);
  static  ChildCode *      Object    (HWND);

  // MDI child window callback

  static  LRESULT CALLBACK MDIChildProc (HWND, UINT, WPARAM, LPARAM);
};

/* ========================================================================== */

FRAMECLASSINFO::FRAMECLASSINFO (HINSTANCE h)
: CLASSINFO (h)
{
  WNDCLASSEX wc;

  wc.cbSize        = sizeof(WNDCLASSEX);
  wc.style         = CS_DBLCLKS;
  wc.lpfnWndProc   = (WNDPROC) FrameCode::MDIFrameProc;
  wc.cbWndExtra    = sizeof (FrameCode *);
  wc.cbClsExtra    = 0;
  wc.hInstance     = hInst;
  wc.hCursor       = LoadCursor (NULL, IDC_ARROW);
  wc.hIcon         = LoadIcon (NULL, IDI_WARNING);
  wc.hbrBackground = (HBRUSH) GetStockObject (LTGRAY_BRUSH);
  wc.lpszMenuName  = NULL;
  wc.lpszClassName = "MDIFrameTest";
  wc.hIconSm       = wc.hIcon;

  ClassAtom = RegisterClassEx (&wc);
};

/* ========================================================================== */

CHILDCLASSINFO::CHILDCLASSINFO (HINSTANCE h)
: CLASSINFO (h)
{
  WNDCLASSEX wc;

  wc.cbSize        = sizeof(WNDCLASSEX);
  wc.style         = CS_DBLCLKS;
  wc.lpfnWndProc   = (WNDPROC) ChildCode::MDIChildProc;
  wc.cbWndExtra    = sizeof (ChildCode *);
  wc.cbClsExtra    = 0;
  wc.hInstance     = hInst;
  wc.hCursor       = LoadCursor (NULL, IDC_ARROW);
  wc.hIcon         = LoadIcon (NULL, IDI_ASTERISK);
  wc.hbrBackground = (HBRUSH) GetStockObject (WHITE_BRUSH);
  wc.lpszMenuName  = NULL;
  wc.lpszClassName = "MDITest";
  wc.hIconSm       = wc.hIcon;

  ClassAtom = RegisterClassEx (&wc);
};

/* ========================================================================== */

FrameCode::FrameCode (HINSTANCE hInst)
: frameclass (hInst),
  childclass (hInst)
{
  hFrame  = NULL;
  hClient = NULL;
  hMenu   = NULL;
};

/* -------------------------------------------------------------------------- */

int FrameCode::Run (void)
{
  if (frameclass.OK() && childclass.OK())
  {
    // Create frame window

    CreateWindow (frameclass.ClassName(),
                  "Test Frame",
                  WS_OVERLAPPEDWINDOW
                  | WS_SYSMENU
                  | WS_CLIPCHILDREN
                  | WS_MINIMIZEBOX
                  | WS_CAPTION
                  | WS_THICKFRAME,
                  CW_USEDEFAULT,
                  CW_USEDEFAULT,
                  800,
                  800,
                  NULL,       // owner
                  NULL,       // menu
                  frameclass.Instance(),
                  (LPVOID) this);

    // Run a message loop

    if (hFrame != NULL && hClient != NULL)
    {
      ShowWindow (hFrame, SW_SHOWDEFAULT);

      MSG msg;

      while (GetMessage (&msg, NULL, 0, 0))
        if (!TranslateMDISysAccel (hClient, &msg))
          if (!IsDialogMessage (msg.hwnd, &msg))
          {
            TranslateMessage (&msg);
            DispatchMessage (&msg);
          };

      return 0;
    };
  };

  return -1;
};

/* -------------------------------------------------------------------------- */

void FrameCode::CreateMenu (void)
{
  hMenu = ::CreateMenu();

  if (hMenu != NULL)
  {
    AppendMenu (hMenu, MF_STRING | MF_ENABLED, -1, "&Run MDI");
    AppendMenu (hMenu, MF_STRING | MF_ENABLED, -2, "&Start MDI");
    AppendMenu (hMenu, MF_STRING | MF_ENABLED, -3, "&Maximize Test");
    AppendMenu (hMenu, MF_STRING | MF_ENABLED, -4, "&Enable/Disable");

    HMENU wlist = CreatePopupMenu();

    AppendMenu (wlist, MF_STRING | MF_ENABLED, ID_TILEVERT, "Tile &Vertically");
    AppendMenu (wlist, MF_STRING | MF_ENABLED, ID_TILEHORZ, "Tile &Horizontally");
    AppendMenu (wlist, MF_STRING | MF_ENABLED, ID_CASCADE,  "&Cascade");
    AppendMenu (wlist, MF_STRING | MF_ENABLED, ID_ARRANGE,  "&Arrange Icons");

    AppendMenu (hMenu, MF_STRING | MF_ENABLED | MF_POPUP, (UINT_PTR) wlist, "&Windows");
    AppendMenu (hMenu, MF_STRING | MF_ENABLED, IDCANCEL, "Close!");

    SendMessage (hClient, WM_MDISETMENU, (WPARAM) hMenu, (LPARAM) wlist);
    DrawMenuBar (hFrame);
  };
};

/* -------------------------------------------------------------------------- */

void FrameCode::RefreshMenu (void) const
{
  SendMessage (hClient, WM_MDIREFRESHMENU, 0, 0);
  DrawMenuBar (hFrame);
};

/* -------------------------------------------------------------------------- */

void FrameCode::OnCommand (WPARAM wParam)
{
  switch ((int) (signed short) LOWORD (wParam))
  {
    case IDCANCEL:
      SendMessage (hFrame, WM_SYSCOMMAND, SC_CLOSE, 0);
      break;
    case -1:
      RunChild (SW_SHOWNORMAL);
      break;
    case -2:
      StartChild();
      break;
    case -3:
      Maximize();
      break;
    case -4:
    {
      HWND active = (HWND) SendMessage (hClient, WM_MDIGETACTIVE, 0, 0);

      if (active != NULL)
      {
        BOOL newstate = !IsWindowEnabled (active);
        EnableWindow (active, newstate);
      };
      break;
    }
    case ID_TILEVERT:
      SendMessage (hClient, WM_MDITILE, MDITILE_VERTICAL, 0);
      break;
    case ID_TILEHORZ:
      SendMessage (hClient, WM_MDITILE, MDITILE_HORIZONTAL, 0);
      break;
    case ID_CASCADE:
      SendMessage (hClient, WM_MDICASCADE, 0, 0);
      break;
    case ID_ARRANGE:
      SendMessage (hClient, WM_MDIICONARRANGE, 0, 0);
      break;
  };
};

/* -------------------------------------------------------------------------- */

void FrameCode::OnClose (void)
{
  if (CloseChildren())
  {
    DestroyMenu (hMenu);
    DestroyWindow (hFrame);
    PostQuitMessage (0);
  }
  else
    RefreshMenu();
};

/* -------------------------------------------------------------------------- */

struct CloseInfo
{
  HWND hFrame;
  HWND hClient;
  HWND active;
};

static BOOL CALLBACK CloseChild (HWND child, LPARAM lParam)
{
  CloseInfo * info = (CloseInfo *) lParam;
  
  assert (info != NULL);

  if (info->active == child || !IsWindow (child) || GetParent (child) != info->hClient)
    return TRUE;

  if (SendMessage (child, WM_CLOSE, 0, 0))
    return FALSE;

  // Check that child window is "our" and kill it

  if (SendMessage (child, XM_NOTIFY, 0, 0))
  {
    MSG msg;

    while (1)
    {
      GetMessage (&msg, info->hFrame, XM_NOTIFY, XM_NOTIFY);

      if (msg.lParam != 0)
      {
        delete (ChildCode *) msg.lParam;
        break;
      };
    };
  };

  return TRUE;
};

BOOL FrameCode::CloseChildren (void)
{
  HWND active = (HWND) SendMessage (hClient, WM_MDIGETACTIVE, 0, (LPARAM) NULL);

  if (active == NULL || !IsWindow (active))
    return TRUE;

  CloseInfo info;

  info.hFrame  = hFrame;
  info.hClient = hClient;
  info.active  = active;

  if (!EnumChildWindows (hClient, CloseChild, (LPARAM) &info))
    return FALSE;

  info.active = NULL;

  return CloseChild (active, (LPARAM) &info);
};

/* -------------------------------------------------------------------------- */

LRESULT CALLBACK FrameCode::MDIFrameProc (HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  FrameCode * frame;

  if (msg == WM_NCCREATE)
  {
    LPCREATESTRUCT ncLPCS = (LPCREATESTRUCT) lParam;

    frame = (FrameCode *) ncLPCS->lpCreateParams;
    frame->Bind (hWnd);
  }
  else
    frame = FrameCode::Object (hWnd);

  switch (msg)
  {
    case WM_NCCREATE:
      frame->hFrame = hWnd;
      break;
    
    case WM_CREATE:
    {
      CLIENTCREATESTRUCT ccs;
      ccs.hWindowMenu = NULL;
      ccs.idFirstChild = 0x100;

      DWORD style = WS_CHILD
                    | MDIS_ALLCHILDSTYLES
                    | WS_CLIPCHILDREN
                    | WS_CLIPSIBLINGS
                    | WS_HSCROLL
                    | WS_VSCROLL;

      frame->hClient = CreateWindowEx (WS_EX_CLIENTEDGE,
                                       "MDICLIENT",
                                       NULL,
                                       style,
                                       0,0,0,0,
                                       hWnd,
                                       (HMENU) 0xCAC,
                                       frame->frameclass.Instance(),
                                       (LPVOID) &ccs);

      frame->CreateMenu();

      BringWindowToTop (frame->hClient);
      ShowWindow (frame->hClient, SW_SHOWNORMAL);
      break;
    }

    case WM_CLOSE:
      frame->OnClose();
      return 0;

    case WM_COMMAND:
      frame->OnCommand (wParam);
      break;

    case WM_NCDESTROY:
      frame->UnBind();
      break;

    case XM_NOTIFY:
      delete (ChildCode *) lParam;
      frame->RefreshMenu();
      return 0;
  };

  if (frame == NULL || frame->hClient == NULL)
    return DefWindowProc (hWnd, msg, wParam, lParam);
  else
    return DefFrameProc (frame->hFrame, frame->hClient, msg, wParam, lParam);
};

/* -------------------------------------------------------------------------- */

void FrameCode::RunChild (int nShowCmd)
{
  ChildCode * child = new ChildCode (this);

  child->Create (nShowCmd);
};

/* -------------------------------------------------------------------------- */

struct CreateChildInfo
{
  FrameCode * frame;
  HWND        hClient;
  int         nShowMode;
};

static HANDLE sema1;                   // waiting by starter
static HANDLE sema2;                   // waiting by launched thread

unsigned _stdcall launch (void *)
{
  // Get a message queue

  MSG msg;
  PeekMessage (&msg, NULL, WM_USER, WM_USER, PM_NOREMOVE);

  ReleaseSemaphore (sema1, 1, NULL);

  // Wait for posting of XM_THREADPARAM

  WaitForSingleObject (sema2, INFINITE);

  // Retrieve create information

  while (!PeekMessage (&msg, (HWND) INVALID_HANDLE_VALUE, XM_THREADPARAM, XM_THREADPARAM, PM_REMOVE))
    Sleep (10);

  // Copy create information

  CreateChildInfo * param = (CreateChildInfo *) msg.lParam;
  CreateChildInfo   info  = *param;

  ReleaseSemaphore (sema1, 1, NULL);

  // Create a MDI child window

  info.frame->RunChild (info.nShowMode);

  // Run a message loop

  while (GetMessage (&msg, NULL, 0, 0))
  {
    if (!TranslateMDISysAccel (info.hClient, &msg))
      if (!IsDialogMessage (msg.hwnd, &msg))
      {
        TranslateMessage (&msg);
        DispatchMessage (&msg);
      };

    if (msg.message == XM_NOTIFY)
      break;
  };

  _endthreadex (0);
  return 0;
};

static void Starter (CreateChildInfo * info)
{
  sema1 = CreateSemaphore (NULL, 0, 1, NULL);
  sema2 = CreateSemaphore (NULL, 0, 1, NULL);

  unsigned  tid;
  _beginthreadex (NULL, 20000, launch, NULL, 0, &tid);

  // Wait while launched thread is creating its message queue

  WaitForSingleObject (sema1, INFINITE);

  PostThreadMessage (tid, XM_THREADPARAM, 0, (LPARAM) info);

  ReleaseSemaphore (sema2, 1, NULL);

  // Wait while launched thread is copying the create information

  WaitForSingleObject (sema1, INFINITE);

  CloseHandle (sema1);
  CloseHandle (sema2);
};

/* -------------------------------------------------------------------------- */

void FrameCode::StartChild (void)
{
  CreateChildInfo info;

  info.frame     = this;
  info.hClient   = hClient;
  info.nShowMode = SW_SHOWNORMAL;

  Starter (&info);
};

/* -------------------------------------------------------------------------- */

void FrameCode::Maximize (void)
{
  CreateChildInfo info;

  info.frame     = this;
  info.hClient   = hClient;
  info.nShowMode = SW_SHOWMAXIMIZED;
  
  int i;

  for (i = 0; i < 64; i++)
    Starter (&info);
};

/* -------------------------------------------------------------------------- */

FrameCode * FrameCode::Object (HWND hWnd)
{
  return (FrameCode *) (LONG_PTR) GetWindowLong (hWnd, 0);
};

/* -------------------------------------------------------------------------- */

void FrameCode::Bind (HWND hWnd)
{
  hFrame = hWnd;
  SetWindowLong (hWnd, 0, (LONG) (LONG_PTR) this);
};

/* -------------------------------------------------------------------------- */

void FrameCode::UnBind (void)
{
  SetWindowLong (hFrame, 0, 0);
};

/* ========================================================================== */

ChildCode::ChildCode (FrameCode * Frame)
{
  assert (Frame != NULL);

  handle = NULL;
  frame  = Frame;
};

/* -------------------------------------------------------------------------- */

void ChildCode::Create (int nShowMode)
{
  char title[32];
  wsprintf (title, "Test Child %08lx  (enabled)", (LONG) (LONG_PTR) this);

  DWORD style = WS_CHILD
                | WS_CLIPSIBLINGS
                | WS_CLIPCHILDREN
                | WS_SYSMENU
                | WS_MINIMIZEBOX
                | WS_MAXIMIZEBOX
                | WS_CAPTION
                | WS_THICKFRAME;

  CreateMDIWindow (frame->childclass.ClassName(),
                   title,
                   style,
                   CW_USEDEFAULT,
                   CW_USEDEFAULT,
                   400,
                   400,
                   frame->hClient,
                   frame->childclass.Instance(),
                   (LPARAM) this);

  if (handle)
  {
    BringWindowToTop (handle);
    ShowWindow (handle, nShowMode);
    SendMessage (frame->hClient, WM_MDIACTIVATE, (WPARAM) handle, 0);

    frame->RefreshMenu();
  };
};

/* -------------------------------------------------------------------------- */

LRESULT CALLBACK ChildCode::MDIChildProc (HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  ChildCode * child;

  switch (msg)
  {
    case WM_NCCREATE:
    {
      LPCREATESTRUCT    ncLPCS  = (LPCREATESTRUCT) lParam;
      LPMDICREATESTRUCT ncLPMCS = (LPMDICREATESTRUCT) ncLPCS->lpCreateParams;

      child = (ChildCode *) ncLPMCS->lParam;
      child->Bind (hWnd);
      break;
    }

    case WM_CLOSE:
      SendMessage (GetParent (hWnd), WM_MDIDESTROY, (WPARAM) hWnd, 0);

      if (IsWindow (hWnd))
        DestroyWindow (hWnd);
      return 0;

    case WM_DESTROY:
      child = ChildCode::Object (hWnd);

      if (child != NULL)
      {
        HWND hFrame = child->frame->hFrame;

        child->UnBind();
        PostMessage (hFrame, XM_NOTIFY, 0, (LPARAM) child);
      };
      return 0;

    case WM_ENABLE:
    {
      MINMAXINFO mmi;
	  SendMessage (hWnd, WM_GETMINMAXINFO, 0, (LPARAM) &mmi);

      RECT clrect;
	  GetWindowRect (GetParent (hWnd), &clrect);
	  MapWindowPoints (NULL, GetParent (hWnd), (LPPOINT) &clrect, 2);

      RECT outer;
	  RECT inner;

	  GetWindowRect (hWnd, &outer);
	  MapWindowPoints (NULL, hWnd, (LPPOINT) &outer, 2);
	  GetClientRect (hWnd, &inner);

      char title[32];
      GetWindowText (hWnd, title, sizeof (title));

      if (!wParam)
        lstrcpy (title + 20, "(disabled)");
      else
        lstrcpy (title + 20, " (enabled)");
      SetWindowText (hWnd, title);

      return 0;
    }

    case WM_NCDESTROY:
      PostThreadMessage (GetWindowThreadProcessId (hWnd, NULL), XM_NOTIFY, 0, 0);
      break;

    case XM_NOTIFY:
      return 1;
  };

  return DefMDIChildProc (hWnd, msg, wParam, lParam);
};

/* -------------------------------------------------------------------------- */

ChildCode * ChildCode::Object (HWND hWnd)
{
  return (ChildCode *) (LONG_PTR) GetWindowLong (hWnd, 0);
};

/* -------------------------------------------------------------------------- */

void ChildCode::Bind (HWND hWnd)
{
  handle = hWnd;

  LONG old = SetWindowLong (handle, 0, (LONG) (LONG_PTR) this);
  assert (old == 0);
};

/* -------------------------------------------------------------------------- */

void ChildCode::UnBind (void)
{
  SetWindowLong (handle, 0, 0);

  handle = NULL;
  frame  = NULL;
};

/* ========================================================================== */

int WINAPI WinMain (HINSTANCE hInst,
                    HINSTANCE,
                    LPSTR,
                    int)
{
  FrameCode frame (hInst);

  return frame.Run();
};

